/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */



#include "../headers/stack.h"

#ifndef RNA_STRUCTURE_H_
#define RNA_STRUCTURE_H_


#define  maxChar 60



/*******************************************/
/*  RNA sequence                           */
/*******************************************/

/* RNA_sequence : array of positions
 * Each position is characterised by
 * - a nucleotide, 
 * - the position of its partner ,
 * - the type of the pairing (-1 for 5prime, 1 for 3prime, or O for free bases)
 */

typedef struct{char nucleotide;int partner;int pairing;}position;
typedef struct{int size;char *name; position * p;}RNA_sequence;


/* CT --> RNA_sequence */
RNA_sequence * get_CT(char * inputFile,char * RNAname);

/*parenthesis --> RNA_sequence */
RNA_sequence * get_bracket(char * inputFile,char * RNAname);


/* display an RNA_sequence in bracket notation */
void display_RNA_sequence_in_bracket_notation(RNA_sequence * rna);




/*******************************************/
/* post-order tree                         */
/*******************************************/

/* post-order tree : array of nodes.
 * Each node is characterised by:
 *  - its size (1 is for leaves)
 *  - its label : labelX is the 5' base, labelY the 3' base. labelY=0 for unpaired bases.  
 *  - its position in the RNA sequence:positionX is the position of the 5'base,and positionY the 
 *    position of the 3' base. baseY=0 for unpaired bases. 
 */

typedef struct{int size;char labelX,labelY;int positionX,positionY;}node;
typedef struct{int size;node * n;}postorder_tree;

/* display a tree structure in post order notation */
void display_postorder_tree(postorder_tree * tree);


int is_subtree(postorder_tree * tree, int i, int j); 

/* RNA_sequence --> post_order */
void RNA_sequence2postorder(RNA_sequence * rna, postorder_tree * tree);



/* RNA_sequence --> triplet post_order */
void RNAsequence2triplet_postorder(RNA_sequence * rna, postorder_tree * tree);

void free_tree(postorder_tree * tree);

void ancestor(postorder_tree * tree, int* tab_ancestor);

/******************************************/
/*   Alignment                            */
/******************************************/

/*post_order --> Parenthesis*/
void alignment(postorder_tree * treeA,postorder_tree * treeB,RNA_sequence *rnaA,RNA_sequence * rnaB, stack * alignment_stack, RNA_sequence * alignA, RNA_sequence * alignB); 

void display_alignment(char * RNAnameA,char * RNAnameB, RNA_sequence * alignA, RNA_sequence * alignB,char * outputFile); 
/* display an alignment (built with the function alignment) */


void create_alignment_RNA_sequences(RNA_sequence *rnaA,RNA_sequence * rnaB,  int **supertree, RNA_sequence * alignA, RNA_sequence * alignB, int startA, int endA, int startB,int endB);	

/* ajout Antoine */
int retrace_base(char *p2, int i);


void pairwise_fasta_output( char *nameA, char *nameB, RNA_sequence *alignA, RNA_sequence *alignB, char *outputFile);



/*********************************/
/*   Miscellaneous               */
/*********************************/

void wrong_format(char * filename); 
int is_RNA_char(char c); 
/* auxiliary function that transforms the encoding into ),(,. for RNA structures*/
char int_2_char(int c); 
int char_2_int(char c); 
char int_to_RNAchar(int a); 

/**
 * control the validity of a nucleotide 
 * and code the nucleotide byan integer 
 *      a = 0                           
 *      c = 1                           
 *      g = 2                           
 *      u = 3                           
 *      - = 4                           
 **/
char RNAcharacter(char c, char * filename); 
int retrace_base(char *p2, int i);
void retrace_partner(int begin, char *p, int *q, int size, char * fileName);

#endif /*RNA_STRUCTURE_H_*/
