/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/********************************/
/*     RNA_structure.c          */
/********************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "../headers/edit_distance.h"
#include "../headers/stack.h"
#include "../headers/score.h"
#include "../headers/RNA_structure.h"
#include "../headers/RNA_multiple_structure.h"

void wrong_format(char * filename){
  printf("Wrong format: %s\n", filename);
  exit(EXIT_SUCCESS);
}

int is_RNA_char(char c){

  c = toupper(c);

  if( c=='T' || c == 'U' || c == 'A' || c == 'G' || c == 'C' )
    return 1;
  else
    return 0;
}

/* auxiliary function that transforms the encoding into ),(,. for RNA structures*/
char int_2_char(int c){
	if(c==-1)return '(';
	if(c==0)return '.';
	if(c==1)return ')';
	if(c==2)return '-';
	return -1;
}

int char_2_int(char c){
  if (c=='(') return -1;
  if (c=='.') return 0;
  if (c==')') return 1;
  wrong_format("  ");
  return -1;
}

char int_to_RNAchar(int a){

  if(a == 0) return 'A';
  else if(a == 1) return 'C';
  else if(a == 2) return 'G';
  else if(a == 3) return 'U';
  else return '-';
}

/**
 * control the validity of a nucleotide 
 * and code the nucleotide byan integer 
 *      a = 0                           
 *      c = 1                           
 *      g = 2                           
 *      u = 3                           
 *      - = 4                           
 **/
char RNAcharacter(char c, char * filename){ 

  if( c == '-' )
    return 4;

  c=toupper(c);
  if (c=='T') c='U';

  if (!((c=='A')||(c=='C')||(c=='G')||(c=='U')))
    wrong_format(filename);

  if(c=='A') return 0;
  else if(c=='C') return 1;
  else if(c=='G') return 2;
  else if(c=='U') return 3;
  else return 5; /* X ou Y*/
}

/**********************************/
/***  DISPLAY FUNCTIONALITIES   ***/
/**********************************/

/* display a tree structure in post order notation */
void display_postorder_tree(postorder_tree * tree){
  int i;
  for(i=1;i<tree->size;i++)
    if(tree->n[i].size==1)
      printf("%d :(%c) %d (%d)\n",i,int_to_RNAchar(tree->n[i].labelX),tree->n[i].size,tree->n[i].positionX);
    else
      printf("%d :(%c-%c) %d (%d,%d)\n",i,int_to_RNAchar(tree->n[i].labelX),int_to_RNAchar(tree->n[i].labelY),
	     tree->n[i].size,tree->n[i].positionX,tree->n[i].positionY);
}

/* display an object with type 'RNA_sequence' in bracket notation */
void display_RNA_sequence_in_bracket_notation(RNA_sequence * rna){
 int i;
 for(i=1;i<rna->size;i++)
   printf("%c", int_to_RNAchar(rna->p[i].nucleotide) );

 printf("\n");
 for(i=1;i<rna->size;i++)
   if (rna->p[i].pairing==-1)
     printf("(");
   else if (rna->p[i].pairing==1)
     printf(")");
   else printf(".");

 printf("\n");
}


 int is_subtree(postorder_tree * tree, int i, int j){
  return ((tree->n[j].size>1) && (j-tree->n[j].size+1==i));
}

void free_tree(postorder_tree * tree){
  free(tree->n);
  free(tree);
}

/*** READING AN RNA STRUCTURE FROM A FILE ***/


/* transform an RNA structure in CT format into an object of type RNA_sequence*/
/* inputFile : name of the file containing the RNA structure in CT format  */
/* RNAname : name of the sequence such as read in the CT file */
RNA_sequence * get_CT(char * inputFile,char * RNAname){
  FILE *fp;
  RNA_sequence * rna;
  int t,i,a,b,c,d,e;
  char ch;
  fp = fopen(inputFile, "rt");
  if(fp==NULL){
    printf("error load : %s\n",inputFile);
    exit(EXIT_SUCCESS);
  }
  fscanf(fp, "%d",&t);/*t is the size of the sequence */
  fgets(RNAname,200,fp); 
  t=t+1;
  rna=(RNA_sequence*) malloc(sizeof(RNA_sequence));
  rna->p = (position *)malloc(sizeof(position)*t);
  rna->size=t;
  
  for(i=1;i<t;i++){
    fscanf(fp, "%d %c %d %d %d %d",&a,&ch,&b,&c,&d,&e);
    /* a is the row number
       ch is the nucleotide (A,C,G,U)
       b=a-1
       c=a+1
       d is the position of the pairing partner. If the current base 
       is an unpaired base, then d=0
       e=a */
    rna->p[i].nucleotide = RNAcharacter(ch, inputFile);
    if (d==0) d=i;
    rna->p[i].partner=d;
    /* if(ch==8)exit(0); helene */
    if(d>i){
      rna->p[i].pairing=-1;
      rna->p[d].pairing=1;
    }
    if(d==i){
      rna->p[i].pairing=0;
    }
  }
  fclose(fp);
  return rna;
}

/*retrouve la parenthse fermante dans une chaine de caractres*/
int retrace_base(char *p2,int i){
  int j;
  if(p2[i]=='('){
    j=1;i++;
    while(j!=0){
      if(p2[i]=='('){j++;i++;}
      else if(p2[i]==')'){j--;i++;}
      else if(p2[i]=='.'){i++;}
    }
    return i-1;
  }else if(p2[i]=='.')return 0;
  else return -1;
}

/*read a file in Vienna bracket format*/
RNA_sequence * get_bracket(char * inputFile,char * RNAname){
  FILE *fp;
  RNA_sequence * rna;
  int size,size1,i,b;
  char *p1;
  char *p2;
  fp = fopen(inputFile, "rt");
  if(fp==NULL){
    printf("error load : %s\n",inputFile);
    exit(EXIT_SUCCESS);
  }
  fgets(RNAname,200,fp);RNAname[0]=' ';
  size1 = ftell(fp);	
  fseek(fp, 0, SEEK_END);
  size = ftell(fp);		
  fseek(fp, size1, SEEK_SET);
  size=size/2;
  p1=malloc(sizeof(char)*size);
  p2=malloc(sizeof(char)*size);
	     
  fscanf(fp,"%s",p1);
  fscanf(fp,"%s",p2);

  size=strlen(p1);
  rna=malloc(sizeof(RNA_sequence));
  rna->p = malloc(sizeof(position)*(size+1));
  rna->size=size+1;

  for(i=0;i<size;i++){
    rna->p[i+1].nucleotide=RNAcharacter(p1[i], inputFile);
    if(p2[i]=='('){
      rna->p[i+1].pairing=-1;
      b=retrace_base(p2,i);
      rna->p[i+1].partner=b+1;
      rna->p[b+1].partner=i+1;
						
    }else if(p2[i]==')'){
      rna->p[i+1].pairing=1;
    }else{
      rna->p[i+1].pairing=0;
      rna->p[i+1].partner=i+1;
    }
  }
  fclose(fp);
  free(p1);
  free(p2);
  return rna;
}

void retrace_partner(int begin, char *p, int *q, int size, char * fileName){
  stack * s=(stack*) malloc(sizeof(stack));
  int i,x,y;

 

  init_stack(s,size);
  for(i=begin; i<size; i++){
      switch(p[i]) {
    case '(' :
      push(s, i, 0);
      break;
    case ')' : 
      if (is_empty_stack(s)) {printf("pile vide !!\n");wrong_format(fileName);}
      pop(s, &x,&y);
      q[i]=x;
      q[x]=i;
      break;
    case '.' : 
      q[i]=i; 
      break;  
    case '-' :
      q[i] = i;
      break;
    }
  }
  if(!is_empty_stack(s)) wrong_format(fileName); 
  free_stack(s);
}


/* ancestor[i]=x : x is the heighest node of a tree
 *  such that
 * i is a leftmost leaf 
 */
void ancestor(postorder_tree * tree, int* tab_ancestor){
  int i ;
 
  for (i=1;i<tree->size; i++){
    tab_ancestor[i]=i;
  }
  for (i=1;i<tree->size; i++){
    tab_ancestor[i-tree->n[i].size+1]=i;
  }
}

/************************************/
/*   CONVERSION  FUNCTIONALITIES    */
/************************************/

/* RNA_sequence --> postorder_tree*/

void RNAsequence_postorder_aux(postorder_tree * tree,RNA_sequence * rna,int x,int y,int * num,int * sizeFather){
  /* auxilliary recursive function - tree is the target postorder_tree, rna is the source RNA_sequence,
      x,y delineate a subsequence in rna, num is the current index in the array for the postorder tree */ 
  int size=1;
  if(x>y)return ;
  if((rna->p[x].partner==y)&& (rna->p[x].pairing!=0)){ /*(x,y) is a base-pair */
    RNAsequence_postorder_aux(tree,rna,x+1,y-1,num,&size);
    *sizeFather=*sizeFather+size;
    *num=*num+1;
    tree->n[*num].labelX=rna->p[x].nucleotide;
    tree->n[*num].labelY=rna->p[y].nucleotide;
    tree->n[*num].size=size;
    tree->n[*num].positionX=x;
    tree->n[*num].positionY=y;
  }else{
    if(rna->p[x].partner==x){  /* x is a free base */
      *num=*num+1;
      size=1;
      *sizeFather=*sizeFather+size;
      tree->n[*num].labelX=rna->p[x].nucleotide;
      tree->n[*num].labelY=0;
      tree->n[*num].size=size;
      tree->n[*num].positionX=x;
      tree->n[*num].positionY=x;
      RNAsequence_postorder_aux(tree,rna,x+1,y,num,sizeFather);			
    }else{   /* x is paired with another base than y */
      RNAsequence_postorder_aux(tree,rna,x,rna->p[x].partner,num,sizeFather);
      RNAsequence_postorder_aux(tree,rna,rna->p[x].partner+1,y,num,sizeFather);
    }
  }
}


void RNA_sequence2postorder(RNA_sequence * rna, postorder_tree * tree){
	int num=0;
	int size=0;
	int i;
	postorder_tree * tmp_tree;
	tmp_tree=malloc(sizeof(postorder_tree));
	tmp_tree->n=malloc(sizeof(node)*(rna->size));
	RNAsequence_postorder_aux(tmp_tree,rna,1,rna->size-1,&num,&size);
	tree->n=malloc(sizeof(node)*(num+1));
	for (i=0; i<num+1; i++){
	  tree->n[i]=tmp_tree->n[i]; 
	}
	tree->size=num+1;
	tree->n[0].positionX=0; /*new*/
	tree->n[0].positionY=0; /*new*/
	free(tmp_tree->n);
	free(tmp_tree);
}




/* bracket --> triplet post_order */
/* each base pair is encoded with one internal node and two leaves */

void RNAsequence_triplet_postorder_aux(postorder_tree * tree,RNA_sequence * rna,int x,int y,int * num,int * sizeFather){
  /* auxilliary recursive function - tree is the target postorder_tree, rna is the source RNA_sequence,
      x,y delineate a subsequence in rna, num is the current index in the array for the postorder tree */ 
  int size=1; 
  if(x>y)return ;
  if ((x<y)&& (rna->p[x].partner==y)){ /*(x,y) is a base-pair */
    *num=*num+1; 
    tree->n[*num].labelX=rna->p[x].nucleotide;
    tree->n[*num].labelY=0;
    tree->n[*num].positionX=x;    
    tree->n[*num].size=1;
    tree->n[*num].positionY=x;
    RNAsequence_triplet_postorder_aux(tree,rna,x+1,y-1,num,&size);
    *sizeFather=*sizeFather+size+2;
    *num=*num+1;
    tree->n[*num].labelX=rna->p[y].nucleotide;
    tree->n[*num].labelY=0;
    tree->n[*num].positionX=y; 
    tree->n[*num].size=1;
    tree->n[*num].positionY=y;
    *num=*num+1;
    tree->n[*num].labelX='X';
    tree->n[*num].labelY='Y';
    tree->n[*num].size=size;
    tree->n[*num].positionX=x;    
    tree->n[*num].positionY=y;
  }
  else{
    if(rna->p[x].pairing==0){ /* x is a free base */
      *num=*num+1;
      size=1;
      *sizeFather=*sizeFather+size;
      tree->n[*num].labelX=rna->p[x].nucleotide;
      tree->n[*num].labelY=0;
      tree->n[*num].size=size;
      tree->n[*num].positionX=x;
      tree->n[*num].positionY=x;
      RNAsequence_triplet_postorder_aux(tree,rna,x+1,y,num,sizeFather);			
    }
    else{ /* x is paired with another base than y */
      RNAsequence_triplet_postorder_aux(tree,rna,x,rna->p[x].partner,num,sizeFather);
      RNAsequence_triplet_postorder_aux(tree,rna,rna->p[x].partner+1,y,num,sizeFather);
    }
  }
}

void RNAsequence2triplet_postorder(RNA_sequence * rna, postorder_tree * tree){
	int num=0;
	int size=0;
	int i;
	postorder_tree * tmp_tree;
	tmp_tree=malloc(sizeof(postorder_tree));
	tmp_tree->n=(node *)malloc(sizeof(node)*(rna->size)*2);
	RNAsequence_triplet_postorder_aux(tmp_tree,rna,1,rna->size-1,&num,&size);
	tree->size=num+1;	
	tree->n=(node *)malloc(sizeof(node)*(num+1));
	for (i=0; i<num+1; i++){
	  tree->n[i]=tmp_tree->n[i]; 
	}

	free(tmp_tree->n);
	free(tmp_tree);
	  /*	tree->n=tmp_tree->n;*/
}


/**************************************/
/*         FINAL  ALIGNMENT           */
/**************************************/



void single2multiple_RNAsequence(RNA_sequence * rna, RNA_sequence_m * rnam){
  
  int i;

  rnam->nb_seq=1; 
  rnam->s = (RNA_sequence *)malloc( sizeof(RNA_sequence) );
  rnam->s[0].size=rna->size;
  rnam->s[0].name=rna->name;
  rnam->s[0].p=rna->p; 
  rnam->super_pairing = (int *)malloc( sizeof(int) * rna->size );
  rnam->super_partner = (int *)malloc( sizeof(int) * rna->size );
  rnam->eq = (char *)malloc( sizeof(char) * rna->size );
  for (i=1; i<rnam->s[0].size; i++){
    rnam->super_partner[i]=rna->p[i].partner;
    rnam->super_pairing[i]=rna->p[i].pairing;
    rnam->eq[i]=0; 
  } 
  rnam->eq[0]=0;
}

/* Create two RNA_sequences that contain the result of the alignment. The mapping is given 
 * in the supertree. The region is delimited by startA..endA and startB..endB 
 */
void create_alignment_RNA_sequences(RNA_sequence *rnaA,RNA_sequence * rnaB,  int **supertree, RNA_sequence * alignA, RNA_sequence * alignB, int startA, int endA, int startB,int endB){

  RNA_sequence_m* alignment, * rnaAm, * rnaBm;
   
  alignment= (RNA_sequence_m*) malloc(sizeof(RNA_sequence_m));

  rnaAm= (RNA_sequence_m*) malloc(sizeof(RNA_sequence_m));
  single2multiple_RNAsequence(rnaA,rnaAm); 

  rnaBm=(RNA_sequence_m*) malloc(sizeof(RNA_sequence_m));
  single2multiple_RNAsequence(rnaB,rnaBm); 

  create_alignment_RNA_sequences_m(rnaAm, rnaBm, supertree, alignment,startA, endA, startB,endB); 
  
  alignA->size= alignment->s[0].size;
  alignA->p=alignment->s[0].p;
  alignB->p=alignment->s[1].p;
  alignB->size= alignment->s[1].size;
}





void display_alignment(char * RNAnameA,char * RNAnameB, RNA_sequence * alignA, RNA_sequence * alignB,char * outputFile){
  int i, j; 
  int alignment_length=alignA->size;
  FILE *fp;
  char * nameA;
  char * nameB; 

  nameA= (char *) malloc(sizeof(char)*15);
  nameB= (char *) malloc(sizeof(char)*15);

  for (i=0; (i<strlen(RNAnameB)) && (i<15) && (RNAnameB[i]!='\n') && (RNAnameB[i]!='\r'); i++){
    nameB[i]=RNAnameB[i];
  } 
  for(j=i; j<15; j++) 
    nameB[j]=' ';

  for (i=0; (i<strlen(RNAnameA)) && (i<15) && (RNAnameA[i]!='\n') && (RNAnameA[i]!='\r'); i++){
    nameA[i]=RNAnameA[i];
  } 
  for(j=i; j<15; j++) 
    nameA[j]=' ';

  if (strlen(outputFile)>0)
    fp = fopen(outputFile, "w");
  else
    fp=stdout;

  for(j=0;j<(alignment_length/maxChar)+1;j++){
    for(i=0; i<15; i++) fprintf(fp,"%c", nameA[i]); fprintf(fp, "   ");
    for(i=(j*maxChar)+1;i<=((j*maxChar)+maxChar) && i<alignment_length;i++){fprintf(fp, "%c",int_2_char(alignA->p[i].pairing));}
    fprintf(fp,"\n");
    for(i=0; i<15; i++) fprintf(fp,"%c", nameA[i]); fprintf(fp,"   ");
    for(i=(j*maxChar)+1;i<=((j*maxChar)+maxChar) && i<alignment_length;i++){fprintf(fp, "%c",int_to_RNAchar(alignA->p[i].nucleotide));}
    fprintf(fp,"\n");
    for(i=0; i<15; i++) fprintf(fp,"%c", nameB[i]); fprintf(fp,"   ");
    for(i=(j*maxChar)+1;i<=((j*maxChar)+maxChar) && i<alignment_length;i++){fprintf(fp, "%c",int_to_RNAchar(alignB->p[i].nucleotide));}
    fprintf(fp,"\n");
    for(i=0; i<15; i++) fprintf(fp,"%c", nameB[i]); fprintf(fp,"   ");
    for(i=(j*maxChar)+1;i<=((j*maxChar)+maxChar) && i<alignment_length;i++){fprintf(fp, "%c",int_2_char(alignB->p[i].pairing));}
    fprintf(fp,"\n");
    fprintf(fp,"                  ");
    for(i=(j*maxChar)+1;i<=((j*maxChar)+maxChar) && i<alignment_length;i++){
      if (alignA->p[i].nucleotide==alignB->p[i].nucleotide) 
	fprintf(fp,"*");
      else
	fprintf(fp," ");
    }
    fprintf(fp,"\n");
    if (j<alignment_length/maxChar) fprintf(fp,"\n");
   
  }
  if (strlen(outputFile)>0)
    fclose(fp);

}


void pairwise_fasta_output( char *nameA, char *nameB, RNA_sequence *alignA, RNA_sequence *alignB, char *outputFile){

  int i;
  FILE *fp;

  if ( strlen(outputFile) > 0 )
    fp = fopen(outputFile, "w");
  else
    fp = stdout;


  /* name A */
  fprintf(fp, ">");
  for(i=0;i<strlen(nameA); i++)
    if(nameA[i] !='\n')
      fprintf(fp, "%c", nameA[i]);
  fprintf(fp, "\n");
  /* nucleotides A */
  for(i=1; i<alignA->size; i++)
    fprintf(fp, "%c", int_to_RNAchar((int)alignA->p[i].nucleotide) );
  fprintf(fp, "\n");
  
  /* brackets A */
  for(i=1; i<alignA->size; i++)
    fprintf(fp, "%c", int_2_char(alignA->p[i].pairing) );
  fprintf(fp, "\n");

  /* name B */
  fprintf(fp, ">");
  for(i=0;i<strlen(nameB); i++)
    if(nameB[i] !='\n')
      fprintf(fp, "%c", nameB[i]);
  fprintf(fp, "\n");
  /* nucleotides B */
  for(i=1; i<alignB->size; i++)
    fprintf(fp, "%c", int_to_RNAchar((int)alignB->p[i].nucleotide) );
  fprintf(fp, "\n");
  
  /* brackets B */
  for(i=1; i<alignB->size; i++)
    fprintf(fp, "%c", int_2_char(alignB->p[i].pairing) );
  fprintf(fp, "\n");

  if( strlen(outputFile) > 0)
    fclose(fp);

}

