/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "../headers/multiple_arc_alignment.h"
#include "../headers/score_m.h"
#include "../headers/stack.h"
#include "../headers/RNA_structure.h"
#include "../headers/RNA_multiple_structure.h"
#include "../headers/graph.h"
#include "../headers/matrix.h"
#include "../headers/constraint.h"
#include "../headers/clustering.h"

static int ** alignment_graph;
static int ** table;
static int * full_deletionA, * full_deletionB;
static char ** is_closedA;
static char ** is_closedB;
static char **constraint;



static int get_table(int x, int y){
  if( constraint[x][y] ==  ALLOWED )
    return table[x][y];
  else
    return -1000000;
}

static void computeTable_Cell_m(postorder_tree_m *treeA, postorder_tree_m *treeB, int x, int y, int i, int j){
  int  d, max_d,w;
  edge ed;
  int w_max = -1000000;

  if (constraint[x][y]==ALLOWED){
  
  /*** we look at all edges coming from (x,y) ***/
  
  /* implicit edges  */
  
  /* full deletion */
  if ( x>i )
    w_max = get_table(x-treeA->n[x].size, y) + full_deletionA[x];
  /* full insertion */
  if ( y>j ){
    w = get_table(x, y-treeB->n[y].size) + full_deletionB[y];
    if ( w>w_max ) { w_max = w; }
  }
  /* substitutions between free bases */
  if ( (x>i) && (y>j) ){  
    w = get_table(x-1, y-1) + base_mismatch_m(treeA->n[x].labelX, treeB->n[y].labelX, treeA->nb_seq, treeB->nb_seq);
    if ( w>w_max ) { w_max = w; }
  }

  table[x][y] = w_max;
  }
}

static void computeTable_m(postorder_tree_m *treeA, postorder_tree_m *treeB, int i, int j, int k, int l){
  int x,y;
  table[i][j] = 0;
  
  for(x=i+1; x<=k; x++)
    table[x][j] = full_deletionA[x] + get_table(x-treeA->n[x].size, j); /*a revoir */
  for(y=j+1; y<=l; y++)
    table[i][y] = full_deletionB[y] + get_table(i, y-treeB->n[y].size); /*a revoir */

  for(x=i+1; x<=k; x++)
    if( is_closedA[i+1][x] )
      for(y=j+1; y<=l; y++)
	if( is_closedB[j+1][y] )
	  computeTable_Cell_m(treeA, treeB, x, y, i, j);	
}

static void build_alignment_graph_m(postorder_tree_m *treeA, postorder_tree_m *treeB){
  int i,j,k,l,m,n,x,y;
  
  alignment_graph[0][0]=0;
  for (j=1; j<treeB->size; j++)
    alignment_graph[0][j]= alignment_graph[0][j-1]+base_deletion_m(treeB->n[j].labelX, treeB->nb_seq, treeA->nb_seq);
  
  for(i=1; i<treeA->size; i++){
    alignment_graph[i][0]=  alignment_graph[i-1][j]+base_deletion_m(treeA->n[i].labelX, treeA->nb_seq, treeB->nb_seq);
    for(j=1; j<treeB->size; j++){
      if (constraint[i][j] == ALLOWED ) { 
	alignment_graph[i][j]=max3(
				   alignment_graph[i-1][j-1]+base_mismatch_m(treeA->n[i].labelX,treeB->n[j].labelX,treeA->nb_seq,treeB->nb_seq),		  alignment_graph[i-1][j]+base_deletion_m(treeA->n[i].labelX, treeA->nb_seq, treeB->nb_seq),  
				   alignment_graph[i][j-1]+base_deletion_m(treeB->n[j].labelX, treeB->nb_seq, treeA->nb_seq));
      }
    }
  }
}


/* Construction of the optimal mapping: 
   the mapping is a set of pairs of nodes of treeA x treeB. 
   It is stored in a stack. */
static int trace_back_alignment_mapping_m(postorder_tree_m *treeA, postorder_tree_m *treeB, int **alignment_graph,
					  stack *mapping){
  int score ;
  int i, j; 
  
  i=treeA->size-1; 
  j=treeB->size-1;
  score=alignment_graph[i][j];
  
  while (i+j>0){
    if ((i*j>0)&&(alignment_graph[i][j] == base_mismatch_m(treeA->n[i].labelX, treeB->n[j].labelX, treeA->nb_seq, treeB->nb_seq) + alignment_graph[-1][y-1])){
      push(mapping, treeA->n[i].positionX, treeB->n[j].positionX);
      i--; j--; 
    }
    else if ((i>0) && (alignment_graph[i][j] == alignment_graph[i-1][j] + base_deletion_m(treeA->n[i].labelX, treeA->nb_seq, treeB->nb_seq))){
      i--; 
    }
    else if ( (j> 0) && (alignment_graph[i][j] == alignment_graph[i][j-1] +  base_deletion_m(treeB->n[j].labelX, treeB->nb_seq, treeA->nb_seq)))
      j--;  
  }
  return score;
}

static int optimal_arc_alignment_m(RNA_sequence_m *rnaA, RNA_sequence_m *rnaB,  RNA_sequence_m *align,
				   int do_mapping, int EQ, int SIMILAR){
  int i, j;
  int score;
  stack *mapping;
  postorder_tree_m *treeA = NULL;
  postorder_tree_m *treeB = NULL;
  
  treeA = RNAsequence2postorder_m(rnaA);
  treeB = RNAsequence2postorder_m(rnaB);
  
  constraint = (char **)calloc(treeA->size, sizeof(char *) );
  for(i=0; i<treeA->size; i++)
    constraint[i] = (char *)calloc( treeB->size, sizeof(char));
  
  alignment_graph= (int**)malloc(sizeof(int *) * treeA->size);
  for(i=0; i<treeA->size; i++)
    alignment_graph[i] = (int *)malloc(sizeof(int) * treeB->size);
  
  no_constraint(constraint, treeA->size, treeB->size);
  
  if( EQ ){
    full_constraint(constraint, treeA->size, treeB->size);
    eq_constraint(constraint, treeA, treeB, rnaA->eq, rnaB->eq, rnaA->s[0].size, rnaB->s[0].size, 1);
  }
  
  if( SIMILAR )
    diagonal_constraint(constraint, treeA->size, treeB->size, (int)((double)( rnaA->s[0].size)/100 * 25) );
  
  
  build_alignment_graph_m(treeA, treeB);
  
  if(do_mapping){
    mapping = (stack *)malloc( sizeof(stack) );
    init_stack(mapping, rnaA->s[0].size); 
    score = trace_back_alignment_mapping_m(treeA, treeB, alignment_graph, mapping, table);
    create_alignment_RNA_sequences_m(rnaA, rnaB, mapping, align, 1, rnaA->s[0].size-1, 1, rnaB->s[0].size-1);
    free_stack(mapping);
  }
  else{
    computeTable_m(treeA, treeB, 0, 0, treeA->size-1, treeB->size-1);
    score = table[treeA->size-1][treeB->size-1];
  }
  
  for(i=0; i<treeA->size; i++){ 
    free(alignment_graph[i]);
    free(constraint[i]);
  }
  
  free(alignment_graph);
  free(constraint);
  
  free_postorder_tree_m(treeA);
  free_postorder_tree_m(treeB);
  free(treeA);
  free(treeB);

  return score;
}

int optimal_multiple_arc_alignment_mapping(RNA_sequence_m *rnaA, RNA_sequence_m *rnaB,  
				    RNA_sequence_m *align, int EQ, int SIMILAR){
  return optimal_arc_alignment_m(rnaA, rnaB, align, 1, EQ, SIMILAR);
}

int optimal_multiple_arc_alignment_score(RNA_sequence_m *rnaA, RNA_sequence_m *rnaB, 
				  int EQ, int SIMILAR){
  RNA_sequence_m *align = NULL; 
  return optimal_arc_alignment_m(rnaA, rnaB, align, 0, EQ, SIMILAR);
}

void multiple_alignment(RNA_sequence_m **list, int nb_seq, int EQ, int SIMILAR, 
			int FASTA, int CLUSTAL, char *nameFileOut ){

  RNA_sequence_m *alignment;
  int score=0, i, j;
  int **tab;
  int max_length=0;
  

  for (i=0; i<nb_seq;i++){
      if (max_length<list[i]->s[0].size) 
	max_length=list[i]->s[0].size; 
  }
  
  tab = (int **)malloc( sizeof(int *) * nb_seq );
  for(i=0; i<nb_seq; i++)
    tab[i] = (int *)malloc( sizeof(int) * (i) );
  
  for(i=0; i < nb_seq; i++)
    for(j=0; j < i; j++)
      tab[i][j] = optimal_multiple_arc_alignment_score(list[i], list[j], EQ, SIMILAR);

  alignment = clustering(tab, nb_seq, list, &score, EQ, SIMILAR);
  
  printf("\nScore : %d", score);  
  printf("\nDensity : %d\n", alignment_density(max_length, nb_seq, alignment));

  if( !FASTA )
    if( !CLUSTAL )
      display_alignment_m(alignment, nameFileOut);
    else
      clustal_w_output(alignment, nameFileOut);
  else
      multi_fasta(alignment, nameFileOut);
  
  /* free memory */
  for(i=0; i<nb_seq; i++)
    free(tab[i]);
  free(tab);
  free_RNA_sequence_m(alignment);
  free(alignment);
}
